@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2010 The QXmpp developers
+ * Copyright (C) 2008-2011 The QXmpp developers
  *
  * Author:
  *  Manjeet Dahiya
@@ -23,49 +23,60 @@
 
 
 #include "QXmppClient.h"
+#include "QXmppClientExtension.h"
+#include "QXmppConstants.h"
 #include "QXmppLogger.h"
-#include "QXmppStream.h"
+#include "QXmppOutgoingClient.h"
 #include "QXmppMessage.h"
-
-#include "QXmppArchiveManager.h"
-#include "QXmppCallManager.h"
-#include "QXmppInvokable.h"
-#include "QXmppMucManager.h"
-#include "QXmppReconnectionManager.h"
-#include "QXmppRpcIq.h"
-#include "QXmppRemoteMethod.h"
-#include "QXmppRosterManager.h"
 #include "QXmppUtils.h"
-#include "QXmppTransferManager.h"
+
+#include "QXmppReconnectionManager.h"
+#include "QXmppRosterManager.h"
 #include "QXmppVCardManager.h"
+#include "QXmppVersionManager.h"
+#include "QXmppEntityTimeManager.h"
+#include "QXmppDiscoveryManager.h"
+#include "QXmppDiscoveryIq.h"
 
 class QXmppClientPrivate
 {
 public:
-    QXmppClientPrivate();
+    QXmppClientPrivate(QXmppClient *);
 
-    QXmppStream* stream;  ///< Pointer to QXmppStream object a wrapper over
+    QList<QXmppClientExtension*> extensions;
+    QXmppLogger *logger;
+    QXmppOutgoingClient* stream;  ///< Pointer to QXmppOutgoingClient object a wrapper over
                           ///< TCP socket and XMPP protocol
     QXmppPresence clientPresence; ///< Stores the current presence of the connected client
 
-    QXmppArchiveManager *archiveManager;  ///< Pointer to the archive manager
-    QXmppCallManager *callManager;        ///< Pointer to the call manager
-    QXmppMucManager *mucManager;          ///< Pointer to the multi-user chat manager
     QXmppReconnectionManager *reconnectionManager;    ///< Pointer to the reconnection manager
     QXmppRosterManager *rosterManager;    ///< Pointer to the roster manager
-    QXmppTransferManager *transferManager;///< Pointer to the transfer manager
     QXmppVCardManager *vCardManager;      ///< Pointer to the vCard manager
+    QXmppVersionManager *versionManager;      ///< Pointer to the version manager
 
-    QHash<QString,QXmppInvokable*> interfaces;
+    void addProperCapability(QXmppPresence& presence);
+
+    QXmppClient *client;
 };
 
-QXmppClientPrivate::QXmppClientPrivate()
+QXmppClientPrivate::QXmppClientPrivate(QXmppClient *parentClient)
     : stream(0),
     clientPresence(QXmppPresence::Available),
-    reconnectionManager(0)
+    reconnectionManager(0), client(parentClient)
 {
 }
 
+void QXmppClientPrivate::addProperCapability(QXmppPresence& presence)
+{
+    QXmppDiscoveryManager* ext = client->findExtension<QXmppDiscoveryManager>();
+    if(ext)
+    {
+        presence.setCapabilityHash("sha-1");
+        presence.setCapabilityNode(ext->clientCapabilitiesNode());
+        presence.setCapabilityVer(ext->capabilities().verificationString());
+    }
+}
+
 /// \mainpage
 ///
 /// QXmpp is a cross-platform C++ XMPP client library based on the Qt
@@ -74,7 +85,8 @@ QXmppClientPrivate::QXmppClientPrivate()
 ///
 /// QXmpp based clients are built using QXmppClient instances which handle the
 /// establishment of the XMPP connection and provide a number of high-level
-/// "managers" to perform specific tasks.
+/// "managers" to perform specific tasks. You can write your own managers to
+/// extend QXmpp by subclassing QXmppClientExtension.
 /// 
 /// <B>Main Class:</B>
 /// - QXmppClient
@@ -86,6 +98,9 @@ QXmppClientPrivate::QXmppClientPrivate()
 /// - QXmppMucManager
 /// - QXmppCallManager
 /// - QXmppArchiveManager
+/// - QXmppVersionManager
+/// - QXmppDiscoveryManager
+/// - QXmppEntityTimeManager
 ///
 /// <B>XMPP stanzas:</B> If you are interested in a more low-level API, you can refer to these
 /// classes.
@@ -108,16 +123,14 @@ QXmppClientPrivate::QXmppClientPrivate()
 /// The default value is 0.
 
 QXmppClient::QXmppClient(QObject *parent)
-    : QObject(parent),
-    d(new QXmppClientPrivate)
+    : QXmppLoggable(parent),
+    d(new QXmppClientPrivate(this))
 {
-    QSslSocket *socket = new QSslSocket;
-    d->stream = new QXmppStream(socket, this);
-    socket->setParent(d->stream);
-    d->clientPresence.setExtensions(d->stream->presenceExtensions());
+    d->stream = new QXmppOutgoingClient(this);
+    d->addProperCapability(d->clientPresence);
 
     bool check = connect(d->stream, SIGNAL(elementReceived(const QDomElement&, bool&)),
-                         this, SIGNAL(elementReceived(const QDomElement&, bool&)));
+                         this, SLOT(slotElementReceived(const QDomElement&, bool&)));
     Q_ASSERT(check);
 
     check = connect(d->stream, SIGNAL(messageReceived(const QXmppMessage&)),
@@ -132,19 +145,15 @@ QXmppClient::QXmppClient(QObject *parent)
         SIGNAL(iqReceived(const QXmppIq&)));
     Q_ASSERT(check);
 
-    check = connect(d->stream, SIGNAL(discoveryIqReceived(const QXmppDiscoveryIq&)), this,
-        SIGNAL(discoveryIqReceived(const QXmppDiscoveryIq&)));
-    Q_ASSERT(check);
-
     check = connect(d->stream, SIGNAL(disconnected()), this,
         SIGNAL(disconnected()));
     Q_ASSERT(check);
 
-    check = connect(d->stream, SIGNAL(xmppConnected()), this,
+    check = connect(d->stream, SIGNAL(connected()), this,
         SLOT(xmppConnected()));
     Q_ASSERT(check);
 
-    check = connect(d->stream, SIGNAL(xmppConnected()), this,
+    check = connect(d->stream, SIGNAL(connected()), this,
         SIGNAL(connected()));
     Q_ASSERT(check);
 
@@ -155,23 +164,34 @@ QXmppClient::QXmppClient(QObject *parent)
     check = setReconnectionManager(new QXmppReconnectionManager(this));
     Q_ASSERT(check);
 
-    // rpc
-    check = connect(d->stream, SIGNAL(rpcCallInvoke(QXmppRpcInvokeIq)),
-        this, SLOT(invokeInterfaceMethod(QXmppRpcInvokeIq)));
-    Q_ASSERT(check);
-
     // logging
-    check = connect(this, SIGNAL(logMessage(QXmppLogger::MessageType, QString)),
-        d->stream, SIGNAL(logMessage(QXmppLogger::MessageType, QString)));
-    Q_ASSERT(check);
+    d->logger = 0;
+    setLogger(QXmppLogger::getLogger());
 
     // create managers
-    d->rosterManager = new QXmppRosterManager(d->stream, this);
-    d->archiveManager = new QXmppArchiveManager(d->stream, this);
-    d->callManager = new QXmppCallManager(d->stream, this);
-    d->mucManager = new QXmppMucManager(d->stream, this);
-    d->transferManager = new QXmppTransferManager(d->stream, this);
-    d->vCardManager = new QXmppVCardManager(d->stream, this);
+    // TODO move manager references to d->extensions
+    d->rosterManager = new QXmppRosterManager(this);
+    addExtension(d->rosterManager);
+
+    d->vCardManager = new QXmppVCardManager;
+    addExtension(d->vCardManager);
+
+    d->versionManager = new QXmppVersionManager;
+    addExtension(d->versionManager);
+
+    addExtension(new QXmppEntityTimeManager());
+
+    QXmppDiscoveryManager *discoveryManager = new QXmppDiscoveryManager;
+    addExtension(discoveryManager);
+
+    // obsolete signal
+    check = connect(discoveryManager, SIGNAL(infoReceived(QXmppDiscoveryIq)),
+                    this, SIGNAL(discoveryIqReceived(QXmppDiscoveryIq)));
+    Q_ASSERT(check);
+
+    check = connect(discoveryManager, SIGNAL(itemsReceived(QXmppDiscoveryIq)),
+                    this, SIGNAL(discoveryIqReceived(QXmppDiscoveryIq)));
+    Q_ASSERT(check);
 }
 
 /// Destructor, destroys the QXmppClient object.
@@ -182,6 +202,50 @@ QXmppClient::~QXmppClient()
     delete d;
 }
 
+/// Registers a new extension with the client.
+///
+/// \param extension
+
+bool QXmppClient::addExtension(QXmppClientExtension* extension)
+{
+    if (d->extensions.contains(extension))
+    {
+        qWarning("Cannot add extension, it has already been added");
+        return false;
+    }
+
+    extension->setParent(this);
+    extension->setClient(this);
+    d->extensions << extension;
+    return true;
+}
+
+/// Unregisters the given extension from the client. If the extension
+/// is found, it will be destroyed.
+///
+/// \param extension
+
+bool QXmppClient::removeExtension(QXmppClientExtension* extension)
+{
+    if (d->extensions.contains(extension))
+    {
+        d->extensions.removeAll(extension);
+        delete extension;
+        return true;
+    } else {
+        qWarning("Cannot remove extension, it was never added");
+        return false;
+    }
+}
+
+/// Returns a list containing all the client's extensions.
+///
+
+QList<QXmppClientExtension*> QXmppClient::extensions()
+{
+    return d->extensions;
+}
+
 /// Returns a modifiable reference to the current configuration of QXmppClient.
 /// \return Reference to the QXmppClient's configuration for the connection.
 
@@ -194,7 +258,7 @@ QXmppConfiguration& QXmppClient::configuration()
 /// are specified using the config parameter. Use signals connected(), error(QXmppClient::Error)
 /// and disconnected() to know the status of the connection.
 /// \param config Specifies the configuration object for connecting the XMPP server.
-/// This contains the host name, user, passwd etc. See QXmppConfiguration for details.
+/// This contains the host name, user, password etc. See QXmppConfiguration for details.
 /// \param initialPresence The initial presence which will be set for this user
 /// after establishing the session. The default value is QXmppPresence::Available
 
@@ -209,11 +273,25 @@ void QXmppClient::connectToServer(const QXmppConfiguration& config,
     }
 
     d->clientPresence = initialPresence;
-    d->clientPresence.setExtensions(d->stream->presenceExtensions());
+    d->addProperCapability(d->clientPresence);
 
     d->stream->connectToHost();
 }
 
+/// Overloaded function to simply connect to an XMPP server with a JID and password.
+///
+/// \param jid JID for the account.
+/// \param password Password for the account.
+
+void QXmppClient::connectToServer(const QString &jid, const QString &password)
+{
+    QXmppConfiguration config;
+    config.setUser(jidToUser(jid));
+    config.setDomain(jidToDomain(jid));
+    config.setPassword(password);
+    connectToServer(config);
+}
+
 /// Overloaded function.
 /// \param host host name of the XMPP server where connection has to be made
 /// (e.g. "jabber.org" and "talk.google.com"). It can also be an IP address in
@@ -221,7 +299,7 @@ void QXmppClient::connectToServer(const QXmppConfiguration& config,
 /// \param user Username of the account at the specified XMPP server. It should
 /// be the name without the domain name. E.g. "qxmpp.test1" and not
 /// "qxmpp.test1@gmail.com"
-/// \param passwd Password for the specified username
+/// \param password Password for the specified username
 /// \param domain Domain name e.g. "gmail.com" and "jabber.org".
 /// \param port Port number at which the XMPP server is listening. The default
 /// value is 5222.
@@ -229,21 +307,17 @@ void QXmppClient::connectToServer(const QXmppConfiguration& config,
 /// after establishing the session. The default value is QXmppPresence::Available
 
 void QXmppClient::connectToServer(const QString& host, const QString& user,
-                                  const QString& passwd, const QString& domain,
+                                  const QString& password, const QString& domain,
                                   int port,
                                   const QXmppPresence& initialPresence)
 {
     QXmppConfiguration &config = d->stream->configuration();
     config.setHost(host);
     config.setUser(user);
-    config.setPasswd(passwd);
+    config.setPassword(password);
     config.setDomain(domain);
     config.setPort(port);
-
-    d->clientPresence = initialPresence;
-    d->clientPresence.setExtensions(d->stream->presenceExtensions());
-
-    d->stream->connectToHost();
+    connectToServer(config, initialPresence);
 }
 
 /// Overloaded function.
@@ -252,7 +326,7 @@ void QXmppClient::connectToServer(const QString& host, const QString& user,
 /// the form of a string (e.g. "192.168.1.25").
 /// \param bareJid BareJid of the account at the specified XMPP server.
 /// (e.g. "qxmpp.test1@gmail.com" or qxmpptest@jabber.org.)
-/// \param passwd Password for the specified username
+/// \param password Password for the specified username
 /// \param port Port number at which the XMPP server is listening. The default
 /// value is 5222.
 /// \param initialPresence The initial presence which will be set for this user
@@ -260,22 +334,17 @@ void QXmppClient::connectToServer(const QString& host, const QString& user,
 
 void QXmppClient::connectToServer(const QString& host,
                                   const QString& bareJid,
-                                  const QString& passwd,
+                                  const QString& password,
                                   int port,
                                   const QXmppPresence& initialPresence)
 {
-    QString user, domain;
-    QStringList list = bareJid.split("@");
-    if(list.size() == 2)
-    {
-        user = list.at(0);
-        domain = list.at(1);
-        connectToServer(host, user, passwd, domain, port, initialPresence);
-    }
-    else
-    {
-        emit logMessage(QXmppLogger::WarningMessage, "Invalid bareJid");
-    }
+    QXmppConfiguration config;
+    config.setHost(host);
+    config.setUser(jidToUser(bareJid));
+    config.setDomain(jidToDomain(bareJid));
+    config.setPassword(password);
+    config.setPort(port);
+    connectToServer(config, initialPresence);
 }
 
 /// After successfully connecting to the server use this function to send
@@ -297,6 +366,19 @@ void QXmppClient::connectToServer(const QString& host,
 
 bool QXmppClient::sendPacket(const QXmppPacket& packet)
 {
+    // Try to cast the packet to Presence packet, and if that succeeds, call the
+    // addProperCapability() on it and send the result, otherwise, if the cast would
+    // fail, the usual sendPacket() would be called.
+    try
+    {
+        // Construct a copy, cause the source parameter is const ref.
+        QXmppPresence pres = dynamic_cast<const QXmppPresence&>(packet);
+        addProperCapability(pres);
+        return d->stream->sendPacket(pres);
+    }
+    catch (...)
+    {
+    }
     return d->stream->sendPacket(packet);
 }
 
@@ -316,8 +398,8 @@ void QXmppClient::disconnectFromServer()
     if (d->stream->isConnected())
     {
         sendPacket(d->clientPresence);
-        d->stream->disconnectFromHost();
     }
+    d->stream->disconnectFromHost();
 }
 
 /// Returns true if the client is connected to the XMPP server.
@@ -339,7 +421,9 @@ QXmppRosterManager& QXmppClient::rosterManager()
 }
 
 /// Utility function to send message to all the resources associated with the
-/// specified bareJid.
+/// specified bareJid. If there are no resources available, that is the contact
+/// is offline or not present in the roster, it will still send a message to
+/// the bareJid.
 ///
 /// \param bareJid bareJid of the receiving entity
 /// \param message Message string to be sent.
@@ -347,9 +431,16 @@ QXmppRosterManager& QXmppClient::rosterManager()
 void QXmppClient::sendMessage(const QString& bareJid, const QString& message)
 {
     QStringList resources = rosterManager().getResources(bareJid);
-    for(int i = 0; i < resources.size(); ++i)
+    if(!resources.isEmpty())
     {
-        sendPacket(QXmppMessage("", bareJid + "/" + resources.at(i), message));
+        for(int i = 0; i < resources.size(); ++i)
+        {
+            sendPacket(QXmppMessage("", bareJid + "/" + resources.at(i), message));
+        }
+    }
+    else
+    {
+        sendPacket(QXmppMessage("", bareJid, message));
     }
 }
 
@@ -393,11 +484,26 @@ void QXmppClient::setClientPresence(const QXmppPresence& presence)
     else
     {
         d->clientPresence = presence;
-        d->clientPresence.setExtensions(d->stream->presenceExtensions());
+        d->addProperCapability(d->clientPresence);
         sendPacket(d->clientPresence);
     }
 }
 
+/// Adds the information about our client to the passed presence packet.
+/// Without this, the packet doesn't contain any information about
+/// client node string, capabilities, etc.
+///
+/// You generally should not need to call this function unless you craft
+/// the presence packets yourself and send them with the sendPacket()
+/// method.
+///
+/// \param presence
+
+void QXmppClient::addProperCapability(QXmppPresence &presence)
+{
+    d->addProperCapability(presence);
+}
+
 /// Function to get reconnection manager. By default there exists a reconnection
 /// manager. See QXmppReconnectionManager for more details of the reconnection
 /// mechanism.
@@ -457,7 +563,7 @@ QXmppStanza::Error::Condition QXmppClient::xmppStreamError()
     return d->stream->xmppStreamError();
 }
 
-/// Returns the reference to QXmppVCardManager, implimentation of XEP-0054.
+/// Returns the reference to QXmppVCardManager, implementation of XEP-0054.
 /// http://xmpp.org/extensions/xep-0054.html
 ///
 
@@ -466,151 +572,50 @@ QXmppVCardManager& QXmppClient::vCardManager()
     return *d->vCardManager;
 }
 
-void QXmppClient::addInvokableInterface( QXmppInvokable *interface )
+/// Returns the reference to QXmppVersionManager, implementation of XEP-0092.
+/// http://xmpp.org/extensions/xep-0092.html
+///
+
+QXmppVersionManager& QXmppClient::versionManager()
 {
-    d->interfaces[ interface->metaObject()->className() ] = interface;
+    return *d->versionManager;
 }
 
+/// Give extensions a chance to handle incoming stanzas.
+///
+/// \param element
+/// \param handled
 
-void QXmppClient::invokeInterfaceMethod( const QXmppRpcInvokeIq &iq )
+void QXmppClient::slotElementReceived(const QDomElement &element, bool &handled)
 {
-    QXmppStanza::Error error;
-    
-    const QStringList methodBits = iq.method().split('.');
-    if (methodBits.size() != 2)
-        return;
-    const QString interface = methodBits.first();
-    const QString method = methodBits.last();
-    QXmppInvokable *iface = d->interfaces.value(interface);
-    if (iface)
+    foreach (QXmppClientExtension *extension, d->extensions)
     {
-        if ( iface->isAuthorized( iq.from() ) )
+        if (extension->handleStanza(element))
         {
-
-            if ( iface->interfaces().contains(method) )
-            {
-                QVariant result = iface->dispatch(method.toLatin1(),
-                                                  iq.arguments() );
-                QXmppRpcResponseIq resultIq;
-                resultIq.setId(iq.id());
-                resultIq.setTo(iq.from());
-                resultIq.setFrom(d->stream->configuration().jid());
-                resultIq.setValues(QVariantList() << result);
-                d->stream->sendPacket( resultIq );
-                return;
-            }
-            else
-            {
-                error.setType(QXmppStanza::Error::Cancel);
-                error.setCondition(QXmppStanza::Error::ItemNotFound);
-
-            }
-        }
-        else
-        {
-            error.setType(QXmppStanza::Error::Auth);
-            error.setCondition(QXmppStanza::Error::Forbidden);
+            handled = true;
+            return;
         }
     }
-    else
-    {
-        error.setType(QXmppStanza::Error::Cancel);
-        error.setCondition(QXmppStanza::Error::ItemNotFound);
-    }
-    QXmppRpcErrorIq errorIq;
-    errorIq.setId(iq.id());
-    errorIq.setTo(iq.from());
-    errorIq.setFrom(d->stream->configuration().jid());
-    errorIq.setQuery( iq );
-    errorIq.setError( error );
-    d->stream->sendPacket( errorIq );
-}
-
-QXmppRemoteMethodResult QXmppClient::callRemoteMethod( const QString &jid,
-                                          const QString &interface,
-                                          const QVariant &arg1,
-                                          const QVariant &arg2,
-                                          const QVariant &arg3,
-                                          const QVariant &arg4,
-                                          const QVariant &arg5,
-                                          const QVariant &arg6,
-                                          const QVariant &arg7,
-                                          const QVariant &arg8,
-                                          const QVariant &arg9,
-                                          const QVariant &arg10 )
-{
-    QVariantList args;
-    if( arg1.isValid() ) args << arg1;
-    if( arg2.isValid() ) args << arg2;
-    if( arg3.isValid() ) args << arg3;
-    if( arg4.isValid() ) args << arg4;
-    if( arg5.isValid() ) args << arg5;
-    if( arg6.isValid() ) args << arg6;
-    if( arg7.isValid() ) args << arg7;
-    if( arg8.isValid() ) args << arg8;
-    if( arg9.isValid() ) args << arg9;
-    if( arg10.isValid() ) args << arg10;
-
-    QXmppRemoteMethod method( jid, interface, args, this );
-    connect( d->stream, SIGNAL(rpcCallResponse(QXmppRpcResponseIq)),
-             &method, SLOT(gotResult(QXmppRpcResponseIq)));
-    connect( d->stream, SIGNAL(rpcCallError(QXmppRpcErrorIq)),
-             &method, SLOT(gotError(QXmppRpcErrorIq)));
-
-
-    return method.call();
-}
-
-/// Returns the reference to QXmppArchiveManager, implementation of XEP-0136.
-/// http://xmpp.org/extensions/xep-0136.html
-///
-
-QXmppArchiveManager& QXmppClient::archiveManager()
-{
-    return *d->archiveManager;
-}
-
-/// Returns the reference to QXmppCallManager, implementation of XEP-0166.
-/// http://xmpp.org/extensions/xep-0166.html
-///
-
-QXmppCallManager& QXmppClient::callManager()
-{
-    return *d->callManager;
-}
-
-/// Returns the reference to QXmppMucManager, implementation of XEP-0045.
-/// http://xmpp.org/extensions/xep-045.html
-///
-QXmppMucManager& QXmppClient::mucManager()
-{
-    return *d->mucManager;
-}
-
-/// Returns the reference to QXmppTransferManager, implementation of:
-///
-///  * XEP-0047: In-Band Bytestreams
-///  * XEP-0095: Stream Initiation
-///  * XEP-0096: SI File Transfer
-///
-
-QXmppTransferManager& QXmppClient::transferManager()
-{
-    return *d->transferManager;
 }
 
 /// Returns the QXmppLogger associated with the current QXmppClient.
 
 QXmppLogger *QXmppClient::logger()
 {
-    return d->stream->logger();
+    return d->logger;
 }
 
 /// Sets the QXmppLogger associated with the current QXmppClient.
 
 void QXmppClient::setLogger(QXmppLogger *logger)
 {
-    d->stream->setLogger(logger);
+    if (d->logger)
+        QObject::disconnect(this, SIGNAL(logMessage(QXmppLogger::MessageType, QString)),
+                   d->logger, SLOT(log(QXmppLogger::MessageType, QString)));
+    d->logger = logger;
+    if (d->logger)
+        connect(this, SIGNAL(logMessage(QXmppLogger::MessageType, QString)),
+                d->logger, SLOT(log(QXmppLogger::MessageType, QString)));
 }
 
 /// At connection establishment, send initial presence.
@@ -620,8 +625,7 @@ void QXmppClient::xmppConnected()
     sendPacket(d->clientPresence);
 }
 
-// obsolete
-
+// deprecated functions
 const QXmppPresence& QXmppClient::getClientPresence() const
 {
     return d->clientPresence;
@@ -691,4 +695,3 @@ QXmppReconnectionManager* QXmppClient::getReconnectionManager()
 {
     return d->reconnectionManager;
 }
-
